@@ -141,7 +141,8 @@ void          bonobo_mdi_set_child_list_path 	(BonoboMDI *mdi,
 gboolean      bonobo_mdi_add_view            	(BonoboMDI *mdi, 
 						 BonoboMDIChild *child);
 gint	      bonobo_mdi_add_views 		(BonoboMDI *mdi, 
-						 GSList *children);
+						 GSList *children,
+						 gboolean set_view);
 gboolean      bonobo_mdi_add_toplevel_view   	(BonoboMDI *mdi, 
 						 BonoboMDIChild *child,
 						 const char *window_role);
@@ -217,7 +218,7 @@ void  bonobo_mdi_set_bottom_pane_for_window (BonoboWindow *win, GtkWidget *pane)
 GtkWidget *bonobo_mdi_get_bottom_pane_for_window (BonoboWindow *win);
 
 gboolean	bonobo_mdi_move_view_to_new_window (BonoboMDI *mdi, GtkWidget *view);
-gint		bonobo_mdi_n_children_for_window (BonoboWindow *win);
+gint		bonobo_mdi_n_views_for_window (BonoboWindow *win);
 
 G_END_DECLS
 
